from pytest import mark

from nyoibo import Entity, fields


def test_additional_value():
    class Model(Entity):
        _int_value = fields.IntField()
        _str_value = fields.StrField()

        def _additional_value(self, key, field, value):
            if isinstance(field, fields.IntField) and value:
                return value + 1
            return value

    instance = Model(int_value=1, str_value='hi')

    assert instance.str_value == 'hi'
    assert instance.int_value == 2


immutable_params = (
    True,
    False
)


@mark.parametrize('immutable', immutable_params)
def test_additional_value_with_different_object_type(immutable):
    class Model0(Entity):
        _value = fields.StrField()

    class Related:

        def __init__(self, instance, field_name):
            self._instance = instance
            self._field_name = field_name

        def fetch(self):
            setattr(self._instance, self._field_name, Model0(value='hi world'))

    class ExtendedLinkField(fields.LinkField):
        _valid_values = (Related,)

    class Model1(Entity):
        _model0 = ExtendedLinkField(to=Model0, immutable=immutable)

        def _additional_value(self, key, field, value):
            if isinstance(field, ExtendedLinkField):
                return Related(self, key)
            return value

    instance0 = Model0(value='hi world')
    instance1 = Model1()

    assert instance0.value == 'hi world'
    assert isinstance(instance1.model0, Related)
    instance1.model0.fetch()
    assert instance1.model0.value == instance0.value



import datetime
from decimal import Decimal

from pytest import mark, raises

from nyoibo import Entity, fields
from nyoibo.exceptions import FieldValueError

str_values = (
    ('10.5', '10.5'),
    (10, '10'),
    (15.2, '15.2')
)


@mark.parametrize('value, expected_result', str_values)
def test_parse_str_value(value, expected_result):
    str_field = fields.StrField()
    assert str_field.parse(value) == expected_result


int_values = (
    ('10', 10),
    (10, 10),
    (15.2, 15),
    ('15.2', 15),
)


@mark.parametrize('value, expected_result', int_values)
def test_parse_int_value(value, expected_result):
    int_field = fields.IntField()
    assert int_field.parse(value) == expected_result


wrong_int_values = (
    'hi',
    datetime.date.today(),
    datetime.datetime.today(),
)


@mark.parametrize('value', wrong_int_values)
def test_wrong_int_value(value):
    int_field = fields.IntField()
    with raises(FieldValueError) as e:
        int_field.parse(value)
    assert str(e.value) == f'{type(value)} is not a valid value for IntField'


bool_values = (
    (True, True),
    (False, False),
    (1, True),
    (0, False),
    ('True', True),
    ('False', False),
    ('false', False),
)


@mark.parametrize('value, expected_result', bool_values)
def test_parse_bool_value(value, expected_result):
    bool_field = fields.BoolField()
    assert bool_field.parse(value) == expected_result


date_values = (
    (datetime.date(2020, 7, 7), datetime.date(2020, 7, 7)),
    ('2020-07-07', datetime.date(2020, 7, 7)),
)


@mark.parametrize('value, expected_result', date_values)
def test_parse_date_value(value, expected_result):
    date_field = fields.DateField()
    assert date_field.parse(value) == expected_result


wrong_date_values = (
    'hi',
    12354
)


@mark.parametrize('value', wrong_date_values)
def test_wrong_date_value(value):
    date_field = fields.DateField()
    with raises(FieldValueError) as e:
        date_field.parse(value)
    assert str(e.value) == f'{type(value)} is not a valid value for DateField'


datetime_values = (
    (
        datetime.datetime(2020, 7, 7, 22, 29, 30),
        datetime.datetime(2020, 7, 7, 22, 29, 30)
    ),
    (
        '2020-07-07T22:29:30',
        datetime.datetime(2020, 7, 7, 22, 29, 30)
    ),
)


@mark.parametrize('value, expected_result', datetime_values)
def test_parse_datetime_value(value, expected_result):
    datetime_field = fields.DatetimeField()
    assert datetime_field.parse(value) == expected_result


wrong_datetime_values = (
    'hi',
    12354
)


@mark.parametrize('value', wrong_datetime_values)
def test_wrong_datetime_value(value):
    datetime_field = fields.DatetimeField()
    with raises(FieldValueError) as e:
        datetime_field.parse(value)
    assert str(e.value) == f'{type(value)} is not a valid value for ' \
                           'DatetimeField'


float_values = (
    (2.5, 2.5),
    (2, 2.0),
    ('2.5', 2.5)
)


@mark.parametrize('value, expected_result', float_values)
def test_parse_float_value(value, expected_result):
    float_field = fields.FloatField()
    assert float_field.parse(value) == expected_result


wrong_float_values = (
    'hi',
    datetime.date.today()
)


@mark.parametrize('value', wrong_float_values)
def test_wrong_float_value(value):
    float_field = fields.FloatField()
    with raises(FieldValueError) as e:
        float_field.parse(value)
    assert str(e.value) == f'{type(value)} is not a valid value for FloatField'


decimal_values = (
    (Decimal('2.5'), Decimal('2.5')),
    ('2.5', Decimal('2.5')),
    (2, Decimal('2')),
)


@mark.parametrize('value, expected_result', decimal_values)
def test_parse_decimal_values(value, expected_result):
    decimal_field = fields.DecimalField()
    assert decimal_field.parse(value) == expected_result


wrong_decimal_values = (
    'hi',
    datetime.date.today()
)


@mark.parametrize('value', wrong_decimal_values)
def test_wrong_decimal_value(value):
    decimal_field = fields.DecimalField()
    with raises(FieldValueError) as e:
        decimal_field.parse(value)
    assert str(e.value) == f'{type(value)} is not a valid value for ' \
                           'DecimalField'


def test_parse_link_field():
    class NewEntity(Entity):
        _value = fields.StrField()

    link_field = fields.LinkField(to=NewEntity)
    new_entity = NewEntity()
    assert link_field.parse(new_entity) == new_entity


def test_parse_none_link_field():
    class NewEntity(Entity):
        pass

    link_field = fields.LinkField(to=NewEntity)
    assert link_field.parse(None) is None


wrong_link_values = (
    '123',
    123,
    datetime.date.today()
)


@mark.parametrize('value', wrong_link_values)
def test_wrong_link_value(value):
    class Example(Entity):
        _value = fields.StrField()

    link_field = fields.LinkField(to=Example)
    with raises(FieldValueError) as e:
        link_field.parse(value)
    assert str(e.value) == f'{type(value)} is not a valid value for LinkField'


def test_parse_link_field_from_dict():
    class NewEntity(Entity):
        _field0 = fields.StrField()
        _field1 = fields.IntField()

    link_field = fields.LinkField(to=NewEntity)
    data = {
        'field0': 'hi',
        'field1': '10'
    }
    parsed = link_field.parse(data)
    assert parsed.field0 == 'hi'
    assert parsed.field1 == 10


def test_parse_link_field_from_dict_with_wrong_data():
    class NewEntity(Entity):
        _field0 = fields.StrField()
        _field1 = fields.IntField()

    link_field = fields.LinkField(to=NewEntity)
    data = {
        'field32': 'hi',
        'field2': '10'
    }
    parsed = link_field.parse(data)
    assert parsed.field0 is None
    assert parsed.field1 is None


def test_parse_dict_field():
    field = fields.DictField()
    data = {
        'name': 'Julian',
        'gender': 'male'
    }
    assert field.parse(data) == data



import datetime
from enum import Enum

from pytest import raises, mark

from nyoibo import Entity, fields
from nyoibo.entities.meta_entity import MetaEntity
from nyoibo.exceptions import PrivateFieldError


class Types(Enum):
    value0 = 'value0'
    value1 = 'value1'
    value2 = 'value2'


class Rate(Entity):
    _value = fields.StrField(immutable=False)
    _other_value = fields.StrField()
    _default = fields.StrField(default_value='hello world')
    _other_default = fields.IntField(default_value=1)
    _private = fields.StrField(private=True)
    _immutable = fields.StrField()
    _type = fields.StrField(choices=Types)


def test_get_value():
    rate = Rate(value='0.25', other_value='hello world')

    assert rate.value == '0.25'
    assert rate.get_value() == '0.25'
    assert rate.other_value == 'hello world'


def test_get_uninitialized_value():
    rate = Rate()

    assert rate.value is None
    assert rate.other_value is None


def test_set_value():
    rate = Rate()
    rate.value = '0.25'

    assert rate.value == '0.25'
    rate.set_value('0.1')
    assert rate.value == '0.1'


def test_fields_must_be_private():
    with raises(PrivateFieldError):
        class Example(Entity):
            value = fields.StrField()


def test_default_values():
    rate = Rate()

    assert rate.default == 'hello world'
    assert rate.other_default == 1


def test_private_field():
    entity = Rate(private='some value')
    with raises(AttributeError):
        print(entity.private)


def test_private_field_getter():
    entity = Rate(private='some value')
    with raises(AttributeError):
        print(entity.get_private())


def test_immutable_field():
    entity = Rate(immutable='some value')

    assert entity.immutable == 'some value'
    with raises(AttributeError):
        entity.immutable = 'other value'

    with raises(AttributeError):
        entity.set_immutable('other value2')


def test_override_getter():
    class Getter(Entity):
        _value = fields.IntField(private=False)

        def get_value(self):
            return self._value + 5

    getter = Getter(value=5)

    assert getter.value == 10
    assert getter.get_value() == 10


def test_override_setter():
    class Setter(Entity):
        _value = fields.IntField(private=False, immutable=False)

        def set_value(self, value):
            self._value = value + 5

    setter = Setter(value=5)

    assert setter.value == 10
    assert setter.get_value() == 10

    setter.value = 10
    assert setter.value == 15
    assert setter.get_value() == 15


def test_override_setter_with_parsing():
    class Setter(Entity):
        _value = fields.IntField(private=False, immutable=False)

        def set_value(self, value):
            value = Setter._value.parse(value)
            self._value = value + 5

    setter = Setter(value=5)

    assert setter.value == 10
    assert setter.get_value() == 10

    setter.value = '10'
    assert setter.value == 15
    assert setter.get_value() == 15


def test_set_value_from_other_field():
    class Setter(Entity):
        _value = fields.StrField(immutable=False)
        _other_value = fields.StrField()

        def set_value(self, value):
            value = Setter._value.parse(value)
            self._value = value
            self._other_value = f'{value} world'

    obj = Setter(value='hi')

    assert obj.value == 'hi'
    assert obj.other_value == 'hi world'


def test_set_value_from_other_field_with_different_order():
    class Setter(Entity):
        _other_value = fields.StrField()
        _value = fields.StrField(immutable=False)

        def set_value(self, value):
            value = Setter._value.parse(value)
            self._value = value
            self._other_value = f'{value} world'

    obj = Setter(value='hi')

    assert obj.value == 'hi'
    assert obj.other_value == 'hi world'


def test_set_none_to_immutable_field_if_value_is_not_passed():
    rate = Rate()
    assert rate.other_value is None


def test_private_value_is_assigned():
    class Private(Entity):
        _add = fields.IntField()
        _value = fields.IntField(private=False)

        def get_value(self):
            return self._add + self._value

    private = Private(add=5, value=5)

    assert private.value == 10


def test_choices():
    entity = Rate(type='value0')

    assert entity.type == Types.value0


wrong_link_to_values = (
    'hi',
    123,
    str,
    int,
    datetime.date
)


@mark.parametrize('to', wrong_link_to_values)
def test_wrong_to_in_link_field(to):
    with raises(ValueError) as e:
        class Example(Entity):
            _value = fields.LinkField(to=to)
    assert str(e.value) == 'to must be an Entity subclass'


def test_link_to_entity_with_other_metaclass():
    class OtherMetaEntity(MetaEntity):
        pass

    class Entity0(Entity, metaclass=OtherMetaEntity):
        _value = fields.StrField()

    class Entity1(Entity):
        _entity0 = fields.LinkField(to=Entity0)


def test_instance_entity_from_dict_with_link_field():
    class Model0(Entity):
        _value = fields.StrField()
        _int_value = fields.IntField()

    class Model1(Entity):
        _link = fields.LinkField(to=Model0)
        _name = fields.StrField()

    data = {
        'name': 'test',
        'link': {
            'value': 'hi world',
            'int_value': '5',
            'different_field': 'some value'
        }
    }
    instance = Model1(**data)
    assert instance.name == 'test'
    assert instance.link.value == 'hi world'
    assert instance.link.int_value == 5


def test_instance_entity_from_dict_with_empty_link_field_data():
    class Model0(Entity):
        _value = fields.StrField()
        _int_value = fields.IntField()

    class Model1(Entity):
        _link = fields.LinkField(to=Model0)
        _name = fields.StrField()

    data = {
        'name': 'test',
        'link': {}
    }
    instance = Model1(**data)
    assert instance.name == 'test'
    assert instance.link is None


def test_instance_entity_from_dict_with_link_field_in_2_depth():
    class Model0(Entity):
        _value = fields.StrField()
        _int_value = fields.IntField()

    class Model1(Entity):
        _link = fields.LinkField(to=Model0)
        _name = fields.StrField()

    class Model2(Entity):
        _value = fields.StrField()
        _link = fields.LinkField(to=Model1)

    data = {
        'value': 'some value',
        'link': {
            'name': 'test',
            'link': {
                'value': 'hi world',
                'int_value': '5',
                'different_field': 'some value'
            }
        }
    }
    instance = Model2(**data)
    assert instance.value == 'some value'
    assert instance.link.name == 'test'
    assert instance.link.link.value == 'hi world'
    assert instance.link.link.int_value == 5



from typing import Dict, Any

from nyoibo.exceptions import PrivateFieldError
from nyoibo.fields import Field


def create_getter(attr):
    def getter(self):
        return getattr(self, attr)
    return getter


def create_setter(attr):
    def setter(self, value):
        field = getattr(self.__class__, attr)
        setattr(self, attr, field.parse(value))
    return setter


class MetaEntity(type):

    def __new__(mcs, name, bases, namespace: Dict[str, Any]):
        getters_setters = {}
        fields = {}
        for attr, value in namespace.items():
            if isinstance(value, Field):
                if not attr.startswith('_'):
                    raise PrivateFieldError('Fields must be private')

                fields[attr] = value
                MetaEntity._set_getters_and_setters(
                    attr,
                    getters_setters,
                    value,
                    namespace
                )

        namespace.update(getters_setters)
        namespace['_fields'] = fields
        return super().__new__(mcs, name, bases, namespace)

    @staticmethod
    def _set_getters_and_setters(attr, getters_setters, field, namespace):
        field_name = attr.replace('_', '', 1)
        if field.private is False:
            getter_name = f'get_{field_name}'
            getter = namespace.get(getter_name) or create_getter(attr=attr)
            getters_setters[getter_name] = getter
            setter = None
            if field.immutable is False:
                setter_name = f'set_{field_name}'
                setter = namespace.get(setter_name) or create_setter(attr=attr)
                getters_setters[setter_name] = setter
            getters_setters[field_name] = property(getter, setter)



import os

from setuptools import setup, Extension, find_packages

try:
    from Cython.Build import cythonize
    USE_CYTHON = True
except ImportError:
    USE_CYTHON = False

if USE_CYTHON:
    ext = 'pyx'
else:
    ext = 'c'

init_file = os.path.join(
    os.path.dirname(__file__),
    'nyoibo',
    '__init__.py'
)

with open(init_file) as f:
    for line in f:
        if line.startswith('__version__ ='):
            _, _, version = line.partition('=')
            VERSION = version.strip(" \n'\"")
            break

with open('README.rst') as f:
    readme = f.read()

tests_require = [
    'pytest==6.2.5'
    'pytest-cov==3.0.0'
]

extensions = [
    Extension('nyoibo.fields', [f'nyoibo/fields.{ext}']),
    Extension('nyoibo.entities.entity', [f'nyoibo/entities/entity.{ext}'])
]

for e in extensions:
    e.cython_directives = {"embedsignature": True}


if USE_CYTHON:
    extensions = cythonize(
        extensions,
        compiler_directives={'language_level': 3}
    )

setup(
    name='nyoibo',
    version=VERSION,
    packages=find_packages(),
    author='Julián Cortés',
    author_email='pity7736@gmail.com',
    description='Implement attributes accessors in an easy way',
    long_description=readme,
    keywords='accessors private',
    ext_modules=extensions,
    url='https://github.com/pity7736/nyoibo',
    tests_require=tests_require,
    python_requires='>=3.7',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Natural Language :: English',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3 :: Only',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Topic :: Software Development',
        'Topic :: Utilities',
        'Typing :: Typed'
    ]
)
